

/////// strformatter /////////////////////////////////////////////////////////:
//
// Format parser
//
// (c) Stephane Rodriguez
//


#include "stdafx.h" // #include <stdio.h> if you are not using precompiled headers
#include "strformatter.h"


strformatter::strformatter()
{
	m_bUseLocalSeparators = FALSE;
}

void strformatter::SetLocale(long lcid)
{
	m_nLCID = lcid;
	m_bUseLocalSeparators = FALSE;
}

void strformatter::SetSeparators(/*in*/char cThousandSep,/*in*/char cDecimalSep)
{
	m_cThousandSep = cThousandSep;
	m_cDecimalSep = cDecimalSep;
	m_bUseLocalSeparators = TRUE;
}


// Methods
//
void strformatter::Format(/*in*/LPSTR szInput, 
						  /*in*/LPSTR szFormat,
						  /*out*/LPSTR szOutputBuffer, /*in*/long nOutputBufferLen)
{
	char cThousandSep = ' '; // default values
	char cDecimalSep = '.';

	// validate input
	if (!szInput || *szInput == 0 || nOutputBufferLen == 0)
	{
		if (szOutputBuffer && nOutputBufferLen > 0)
			*szOutputBuffer = 0;

		return;
	}

	//
	// validate format
	//
	BOOL bNotOnlySpaces = FALSE;
	if (szFormat)
	{
		long n = strlen(szFormat);
		long i = 0;
		while (i < n && !bNotOnlySpaces)
			bNotOnlySpaces = (szFormat[i++] != ' ');
	}

	if (!szFormat || *szFormat == 0 || strstr(szFormat,"Standard") != NULL || !bNotOnlySpaces)
	{
		if (szOutputBuffer && nOutputBufferLen > 0)
		{
			// pass the input as is
			long n = strlen(szInput);
			if (n < nOutputBufferLen-1)
				strcpy(szOutputBuffer, szInput);
			else
			{
				strncpy(szOutputBuffer, szInput, nOutputBufferLen-1);
				szOutputBuffer[nOutputBufferLen-1] = 0;
			}
		}
		
		return;
	}

	//
	// retrieve separators from current locale
	//
	if ( !m_bUseLocalSeparators )
	{
		TCHAR szTmp[256]="";
		if ( ::GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_STHOUSAND, szTmp, 256) )
			cThousandSep = *szTmp;

		if ( ::GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_SDECIMAL, szTmp, 256) )
			cDecimalSep = *szTmp;
	}
	else
	{
		cThousandSep = m_cThousandSep;
		cDecimalSep = m_cDecimalSep;
	}

	//
	// parse format
	//

	*szOutputBuffer = 0;

	LPSTR szPtrDot = strchr(szFormat, cDecimalSep);

	FormatLeftHand(szInput, szFormat, szPtrDot, szOutputBuffer, nOutputBufferLen, cThousandSep, cDecimalSep);
	FormatRightHand(szInput, szFormat, szPtrDot, szOutputBuffer, nOutputBufferLen, cThousandSep, cDecimalSep);
}


void strformatter::FormatLeftHand(/*in*/LPSTR szInput, /*in*/LPSTR szFormat, /*in*/LPSTR szPtrDot,
								  /*out*/LPSTR szOutputBuffer, /*in*/long nOutputBufferLen,
								  /*in*/char cThousandSep,
								  /*in*/char cDecimalSep)
{

	long ninput = strlen(szInput);

	LPSTR pinput = strchr(szInput, cDecimalSep);
	if (!pinput)
		pinput = szInput + ninput; // last character of input
	pinput--;

	LPSTR p = szPtrDot;
	if ( !p )
	{
		long nformat = strlen(szFormat);
		p = szFormat + nformat;
	}
	p--;

	LPSTR poutput = szOutputBuffer;
	BOOL bWithinDigits = FALSE;

	// now parse the string right to left
	while ( (p >= szFormat) && (poutput-szOutputBuffer < nOutputBufferLen-1) )
	{
		// do we have a wild card (0 or #)? or a space? or anything else?
		if (*p == '0')
		{
			if (pinput >= szInput && IsDigit(pinput) )
				*(poutput++) = *(pinput--);
			else
				*(poutput++) = '0';

			bWithinDigits = TRUE;
		}
		else if (*p == '#' && (pinput >= szInput) && IsDigit(pinput) )
		{
			*(poutput++) = *(pinput--);
			bWithinDigits = TRUE;
		}
		else if (*p == cThousandSep)
		{
			if ( pinput >= szInput ) // add a separator (only if there are remaining things in the input)
				*(poutput++) = cThousandSep;

			// pass all thousand separators in the input
			while ( (pinput >= szInput) && (*pinput == cThousandSep) )
				pinput--;
		}
		else
		{
			// just before we output the non wildchard character, just output the remaining input digits
			if ( bWithinDigits )
			{
				while ( (pinput >= szInput) && (poutput-szOutputBuffer < nOutputBufferLen-1) )
					*(poutput++) = *(pinput--);

				bWithinDigits = FALSE; // loop to allow patterns of the form A##B##C (even if they are weird)
			}

			if ( (poutput-szOutputBuffer < nOutputBufferLen-1) )
				*(poutput++) = *p;
		}

		p--;
	}
	if (pinput >= szInput) // add remaining chars if any
	{
		while ( (pinput >= szInput) && (poutput-szOutputBuffer < nOutputBufferLen-1) )
			*(poutput++) = *(pinput--);
	}

	// reverse the output
	*(poutput) = 0;
	strrev(szOutputBuffer);

	// add the decimal
	if (szPtrDot && (poutput - szOutputBuffer < nOutputBufferLen-1) )
	{
		*(poutput++) = cDecimalSep;
		if (poutput - szOutputBuffer < nOutputBufferLen-1)
			*(poutput++) = 0;
		else 
			*(poutput - 1) = 0; // override the last character
			
	}

}

void strformatter::FormatRightHand(/*in*/LPSTR szInput, /*in*/LPSTR szFormat, /*in*/LPSTR szPtrDot,
								   /*out*/LPSTR szOutputBuffer, /*in*/long nOutputBufferLen,
								   /*in*/char cThousandSep,
								   /*in*/char cDecimalSep)
{
	// 
	long noutput = strlen(szOutputBuffer);
	if (noutput >= nOutputBufferLen) return;

	LPSTR poutput = szOutputBuffer + noutput;

	if (!szPtrDot) return; // there is no decimal, so there's nothing left to do

	LPSTR p = szPtrDot;

	// pass the decimals
	while ( *p != 0 && *p == cDecimalSep)
		p++;

	if (*p == 0)
		return; // the remainder of the input is lost

	long ninput = strlen(szInput);

	LPSTR pinput = strchr(szInput, cDecimalSep);
	if (!pinput)
		pinput = szInput + ninput; // last character of input
	else
		pinput++;

	// the main difference with left hand, is that right hand width is limited
	while ( *p != 0 && (poutput-szOutputBuffer < nOutputBufferLen-1) )
	{
		// do we have a wild card (0 or #)? or a space? or anything else?
		if (*p == '0')
		{
			if ( (pinput-szInput < ninput) && IsDigit(pinput) )
				*(poutput++) = *(pinput++);
			else
				*(poutput++) = '0';

		}
		else if (*p == '#' && (pinput-szInput < ninput) && IsDigit(pinput) )
		{
			*(poutput++) = *(pinput++);
		}
		else if (*p == cThousandSep)
		{
			*(poutput++) = cThousandSep;

			// pass all thousand separators in the input
			while ( (pinput-szInput < ninput) && (*pinput == cThousandSep) )
				pinput++;
		}
		else
		{
			if ( (poutput-szOutputBuffer < nOutputBufferLen-1) )
				*(poutput++) = *p;
		}

		p++;
	} // end while

	*poutput = 0;
}


BOOL strformatter::IsDigit(LPSTR p)
{
	if ( !p ) return FALSE;

	return ( *p >= '0' && *p <= '9');
}